#
#  Copyright 2024 by Dmitry Berezovsky, MIT License
#
import abc
from functools import wraps
import importlib
import importlib.util
import inspect
import logging
import typing
from typing import Any, Callable, Generic, Sequence, TypeAlias, TypeVar

import injector
from injector import T

from unikit.dto.interfaces import Bootstrapable, Nameable
from unikit.registry import Registry

logger = logging.getLogger(__name__)

# Types
IT = TypeVar("IT")
BindArg: TypeAlias = IT | Callable[..., IT] | injector.Provider[IT] | type[IT]


class DIContainerSetupError(Exception):
    """Error raised when the DI container setup fails."""

    pass


class DiPlaceholder:
    """Placeholder for the DI container to inject the value."""

    def __getattribute__(self, item: Any) -> Any:
        if item in ("__class__", "__deepcopy__", "__reduce_ex__", "__reduce__", "__getstate__"):
            return super().__getattribute__(item)
        raise RuntimeError(f"This parameter must be injected by DI container (invoked attribute: {item})")

    def __str__(self) -> str:
        return "Should be injected by DI container"

    def __repr__(self) -> str:
        return self.__str__()


PLACEHOLDER: Any = DiPlaceholder()


class InterfaceAwareProvider(injector.Provider[T], Generic[T], abc.ABC):
    """Base class for providers that are aware of the interface they provide."""

    def __init__(self, interface_cls: type[T] | None) -> None:
        self.iface_cls = interface_cls


class ClassListProvider(InterfaceAwareProvider[list[T]], Generic[T]):
    """Provides a list of instances from a given class."""

    def __init__(self, cls: type[T], interface_cls: type[list[T]] | None) -> None:
        """Initialize the provider."""
        self._cls = cls
        super().__init__(interface_cls=interface_cls)

    def get(self, injector: injector.Injector) -> list[T]:
        """Get the list of instances."""
        return [injector.create_object(self._cls)]


class InterfaceAwareClassProvider(InterfaceAwareProvider[T], injector.ClassProvider, Generic[T]):
    """Provides a class instance with the given interface."""

    def __init__(self, cls: type[T], interface_cls: type[T] | None) -> None:
        """Initialize the provider."""
        self._cls = cls
        super().__init__(interface_cls=interface_cls)

    def get(self, injector_: injector.Injector) -> T:  # type: ignore
        """Get the class instance."""
        return injector_.create_object(self._cls)


class MultiBindClassProvider(injector.MultiBindProvider):
    """A provider for a list of instances from a list of classes."""

    def __init__(self, classes: Sequence[type[T]], interface_cls: type[list[T]] | None) -> None:
        super().__init__()
        for cls in classes:
            self.append(ClassListProvider(cls, interface_cls))


class IfaceAwareSingletonScope(injector.SingletonScope):
    """A singleton scope that is aware of the interface of the provided object."""

    @injector.synchronized(injector.lock)
    def get(self, key: type[T], provider: injector.Provider[T]) -> injector.Provider[T]:
        """Get the provider."""

        if key in self._context:
            return self._context[key]
        elif (
            isinstance(provider, InterfaceAwareProvider) and provider.iface_cls and provider.iface_cls in self._context
        ):
            return self._context[provider.iface_cls]
        instance_provider = injector.InstanceProvider(provider.get(self.injector))
        self._context[key] = instance_provider
        if isinstance(provider, injector.MultiBindProvider) and isinstance(instance_provider._instance, list):  # noqa
            for x in instance_provider._instance:  # noqa
                self._context[x.__class__] = injector.InstanceProvider(x)
        if isinstance(provider, InterfaceAwareProvider) and provider.iface_cls:
            self._context[provider.iface_cls] = instance_provider
        return instance_provider


class DiModule(injector.Module):
    """Base class for DI modules."""

    def __init__(self, binder: injector.Binder | None = None) -> None:
        """Initialize the module."""
        super().__init__()
        self.__binder: injector.Binder | None = binder
        if binder:
            super().__call__(binder)

    def configure(self, binder: injector.Binder) -> None:
        """Configure DI container."""
        super().configure(binder)
        self.__binder = binder

    def register_singleton(self, interface: type, to: BindArg | None = None) -> None:
        """
        Register given interface and optionally implementation in DI container with SingletonScope.

        The instance will be created only once per DI container at the moment of first access.

        Skipped `to` argument is equivalent to `register_singleton(ClassA, ClassA)` and must be used in cases when you
        do not have a dedicated interface for the implementation (generally a bad practice).

        Important note: subsequent invocations with the same `interface` argument will __override__ previously set
        implementation.

        :param interface: interface to be registered
        :param to: implementation to be associated with given interface.
        """

        assert self.__binder is not None
        # noinspection PyTypeChecker
        self.__binder.bind(interface, to=to, scope=IfaceAwareSingletonScope)
        if interface and to and interface is not to and isinstance(to, type):
            # noinspection PyTypeChecker
            self.__binder.bind(
                to,
                to=InterfaceAwareClassProvider(to, interface_cls=interface),
                scope=IfaceAwareSingletonScope,
            )

    def register_singleton_impl(self, interface: type, to: Sequence[BindArg] | BindArg) -> None:
        """
        Similar to `register_singleton` but allows to register *multiple* implementations of the same interface.

        To access implementations you __must__ inject a `List` of interfaces, not a single interface.
        """
        assert self.__binder is not None
        if not isinstance(to, (list, tuple)):
            to = [to]
        self.__binder.multibind(
            list[interface],  # type: ignore
            MultiBindClassProvider(to, interface_cls=interface),  # type: ignore
            scope=IfaceAwareSingletonScope,
        )
        for x in to:
            if isinstance(x, type):
                self.register_singleton(interface=x, to=InterfaceAwareClassProvider(x, interface_cls=x))

    def add_to_registry(
        self, registry_cls: type[Registry], obj: BindArg, key: Any = None, skip_container: bool = False
    ) -> None:
        """Add given object to the registry."""
        assert self.__binder is not None
        assert not isinstance(registry_cls, Registry), "Registry class must be a subclass of Registry"
        interface = self.__get_registry_interface(registry_cls)
        try:
            self.__binder.get_binding(registry_cls)
        except injector.UnsatisfiedRequirement:
            self.register_singleton(registry_cls)
        registry = self.__binder.injector.get(registry_cls)
        if not skip_container:
            if interface is not None:
                self.register_singleton_impl(interface, obj)
            else:
                self.register_singleton(obj)  # type: ignore
        if key is None and isinstance(obj, Nameable) or (isinstance(obj, type) and issubclass(obj, Nameable)):
            registry.register(obj.get_name(), obj)
        else:
            if key is None:
                raise ValueError("Key must be provided")
            registry.register(key, obj)

    def __get_registry_interface(self, registry_cls: type[Registry]) -> type | None:
        for parent in registry_cls.mro():
            if not issubclass(parent, Registry):
                continue
            for base_class in getattr(parent, "__orig_bases__", ()):
                type_args = typing.get_args(base_class)
                if len(type_args) == 2:
                    return type_args[1]
                elif len(type_args) == 1:
                    return type_args[0]
        return None


class DiInjector(injector.Injector):
    """Extended injector with additional functionality."""

    def autodiscover_modules(self, packages: Sequence[str] | str) -> None:
        """Autodiscover modules in given packages and install them in DI container."""
        if isinstance(packages, str):
            packages = (packages,)
        for target_package_name in packages:
            self.__discover_module(target_package_name)

    def __discover_module(self, target_package_name: str) -> None:
        spec = importlib.util.find_spec(".di_context", target_package_name)
        if spec:
            try:
                module = importlib.import_module(spec.name)
                for name, obj in inspect.getmembers(module, inspect.isclass):
                    if issubclass(obj, DiModule) and obj is not DiModule and not inspect.isabstract(obj):
                        logger.info("Installing DI module %s at %s", name, spec.name)
                        self.binder.install(obj)
            except Exception as e:
                raise DIContainerSetupError(f"Unable to bootstrap DI container for {target_package_name}") from e

    def find_boostrapable(self) -> list[type[Bootstrapable]]:
        """Find all bootstrapable classes in DI container."""
        result: list[type[Bootstrapable]] = []
        for _, bindings in self.binder._bindings.items():
            for x in bindings:
                try:
                    if inspect.isclass(x) and issubclass(x, Bootstrapable):
                        result.append(x)
                except TypeError:
                    pass
        return result


class Initializable(abc.ABC):
    """
    Interface for objects that need to be initialized after creation.

    This might be useful if you have to create class instance before without DI context and have to inject
    dependencies later.
    """

    def __init__(self) -> None:
        super().__init__()
        self.__initialized = False

    def initialize(self, injector_: injector.Inject[injector.Injector]) -> None:
        """Initialize object."""
        injector_.call_with_injection(self._do_initialize)
        self.__initialized = True

    @abc.abstractmethod
    def _do_initialize(self, **kwargs: Any) -> None:
        pass

    @property
    def is_initialized(self) -> bool:
        """Return true if object is initialized."""
        return self.__initialized


root_container = DiInjector(auto_bind=False)
root_container.binder.bind(DiInjector, injector.InstanceProvider(root_container), scope=injector.SingletonScope)


def with_injection(fn: Callable[..., Any]) -> Callable[..., Any]:
    """
    Wrap given function with root_container.call_with_injection.

    INTENDED TO BE USED IN TESTS ONLY!
    :param fn: target function
    """

    @wraps(fn)
    def wrapper(*args: Any, **kwargs: Any) -> Any:
        return root_container.call_with_injection(fn, args=args, kwargs=kwargs)

    return wrapper


def has_inject_marker(v: type) -> bool:
    """
    Return true if given type/type annotation is marked for injection.

    :param v: type to test
    :return: true if injector marker is present
    """
    return (
        injector._is_specialization(v, typing.Annotated)
        and injector._inject_marker in v.__metadata__  # type: ignore[attr-defined]
    )


class DiProxy(Generic[T]):
    """Proxy object that delegates all calls to the underlying object."""

    _OWN_PROPERTIES = ("_proxy_di_provider", "_proxy_injector")

    def __init__(self, di_provider: injector.Provider[T], injector: injector.Injector) -> None:
        self._proxy_di_provider = di_provider
        self._proxy_injector = injector

    @property
    def underlying_object(self) -> T:
        """Return the underlying object."""
        return self._proxy_di_provider.get(self._proxy_injector)

    def __getattr__(self, item: str) -> Any:
        if item in DiProxy._OWN_PROPERTIES:
            return super().__getattr__(item)  # type: ignore
        return getattr(self.underlying_object, item)

    def __setattr__(self, key: str, value: Any) -> None:
        if key in DiProxy._OWN_PROPERTIES:
            return super().__setattr__(key, value)
        return setattr(self.underlying_object, key, value)

    def __delattr__(self, item: str) -> None:
        if item in DiProxy._OWN_PROPERTIES:
            return super().__delattr__(item)
        return delattr(self.underlying_object, item)

    def __dir__(self) -> typing.Iterable[str]:
        return dir(self.underlying_object)

    def __str__(self) -> str:
        return f"Proxy<{str(self.underlying_object)}>"

    def __hasattr__(self, item: str) -> bool:
        return hasattr(self.underlying_object, item)

    def __eq__(self, other: Any) -> bool:
        return self.underlying_object == other



#
#  Copyright 2024 by Dmitry Berezovsky, MIT License
#
import abc
import logging

from django.apps import AppConfig

from unikit.di import root_container

logger = logging.getLogger(__name__)


class BaseDiSupportedApp(AppConfig, abc.ABC):  # type: ignore[misc]
    """Base class for applications that support Dependency Injection."""

    default = False

    def __init__(self, app_name: str, app_module: str) -> None:
        super().__init__(app_name, app_module)
        self._di_container = root_container

    def _install_di_modules(self) -> None:
        """Scan current application for DI Module declaration and install discovered modules into root container."""
        logger.debug("Discovering DI modules for app %s", self.name)
        target_package_name = ".".join(self.__module__.split(".")[:-1])
        self._di_container.autodiscover_modules(target_package_name)

    def ready(self) -> None:
        """Django callback invoked when application is ready to be used."""
        super().ready()
        # Import DI context if any
        self._install_di_modules()



#
#  Copyright 2024 by Dmitry Berezovsky, MIT License
#
from typing import Mapping, TypeVar, cast

TDict = TypeVar("TDict", bound=Mapping)


def deepmerge(to_dict: TDict, from_dict: TDict, merge_lists: bool = False) -> TDict:
    """
    Merge two dictionaries recursively and return created dict. Dict `from_dict` will be merged into dict `to_dict`.

    This is a PURE function (doesn't have side effects), so it will not modify the original dictionaries.
    If a key is present in both dictionaries, the value from `new` will be used.

    :param to_dict: original object which will be used as a base
    :param from_dict: new object which will be merged into original
    :param merge_lists: whether to merge lists or replace them
    """

    def merge(_old: TDict, _new: TDict) -> TDict:
        # pylint: disable=no-else-return
        if isinstance(_new, dict):
            if not isinstance(_old, dict):
                return cast(TDict, _new)
            res = _old.copy()
            for k, v in _new.items():
                res[k] = merge(_old[k], v) if k in _old else v
            return cast(TDict, res)
        elif isinstance(_new, list):
            if merge_lists:
                if not isinstance(_old, list):
                    return _new
                return _old + _new
            else:
                return _new
        return _new

    return merge(to_dict, from_dict)



#
#  Copyright 2024 by Dmitry Berezovsky, MIT License
#
import datetime
from typing import Any, Protocol, TypeVar

T = TypeVar("T")

Number = int | float
TAnyDate = TypeVar("TAnyDate", bound=datetime.date)

_T_contra = TypeVar("_T_contra", contravariant=True)


class SupportsDunderLT(Protocol[_T_contra]):
    """A type that supports the `<` operator."""

    def __lt__(self, __other: _T_contra) -> bool: ...


class SupportsDunderGT(Protocol[_T_contra]):
    """A type that supports the `>` operator."""

    def __gt__(self, __other: _T_contra) -> bool: ...


class SupportsDunderLE(Protocol[_T_contra]):
    """A type that supports the `<=` operator."""

    def __le__(self, __other: _T_contra) -> bool: ...


class SupportsDunderGE(Protocol[_T_contra]):
    """A type that supports the `>=` operator."""

    def __ge__(self, __other: _T_contra) -> bool: ...


class SupportsAllComparisons(
    SupportsDunderLT[Any], SupportsDunderGT[Any], SupportsDunderLE[Any], SupportsDunderGE[Any], Protocol
):
    """A type that supports all comparison operators."""

    ...


TSupportsAllComparisons = TypeVar("TSupportsAllComparisons", bound=SupportsAllComparisons)


def none_raises(optional: T | None, message: str = "Unexpected `None`") -> T:
    """Convert an optional to its value."""
    if optional is None:
        raise AssertionError(message)
    return optional



#
#  Copyright 2024 by Dmitry Berezovsky, MIT License
#
import logging


class LogMixin:
    """Convenience super-class to have a logger configured with the class name."""

    _log: logging.Logger | None = None

    @property
    def log(self) -> logging.Logger:
        """Return a logger."""
        if self._log is None:
            self._log = logging.getLogger(self.__class__.__module__ + "." + self.__class__.__name__)
        return self._log



#
#  Copyright 2024 by Dmitry Berezovsky, MIT License
#
import abc
from hashlib import sha1
from typing import Any


class Nameable(abc.ABC):
    """Represent the object which has a name."""

    @classmethod
    @abc.abstractmethod
    def get_name(cls) -> str:
        """Get the name of the object."""
        pass


class Hashable(abc.ABC):
    """Represent the object which can be hashed."""

    @abc.abstractmethod
    def get_hash(self) -> str:
        """Calculate hash for this object."""
        pass


class HashableHelper(Hashable, metaclass=abc.ABCMeta):
    """Helper class to implement Hashable interface."""

    def get_hash(self) -> str:
        """Calculate hash for this object."""
        return sha1(self._get_hashable_input().encode("utf-8")).hexdigest()

    @abc.abstractmethod
    def _get_hashable_input(self) -> str:
        """Get the input for hash calculation."""
        pass


class Bootstrapable(abc.ABC):
    """Represent the object which requires some heavy initialization before it can be used."""

    @classmethod
    @abc.abstractmethod
    def bootstrap(cls, *args: Any, **kwargs: Any) -> bool:
        """
        Bootstrap the object. This method is called in build time to ensure all required resources are loaded.

        :return: True if bootstrap has been performed, False if bootstrap is not required
        :raises: Exception if bootstrap has failed
        """
        pass



#
#  Copyright 2024 by Dmitry Berezovsky, MIT License
#
__all__ = ("raise_or_default", "OnErrorDef")

from typing import TypeVar, Union, cast

TDefault = TypeVar("TDefault")
OnErrorDef = Union[type[Exception], TDefault, None]


def raise_or_default(on_error: OnErrorDef[TDefault], error_msg: str) -> TDefault | None:
    """Raise an exception or return a default value."""
    if isinstance(on_error, type) and issubclass(on_error, Exception):
        raise on_error(error_msg)
    return cast(TDefault, on_error)



#
#  Copyright 2024 by Dmitry Berezovsky, MIT License
#
import hashlib
from io import IOBase
import mimetypes
import os
import random
import string
from typing import IO, Literal, TextIO, TypeVar, Union, cast

CONTENT_TYPE_GENERIC = "application/octet-stream"
FILE_NAME_ALPHABET = string.ascii_lowercase + string.digits


def get_ext_from_content_type(content_type: str) -> str:
    """
    Get file extension from content type.

    :param content_type: content type
    :return: file extension with dot. E.g. `.txt`
    """
    if content_type == "application/json":
        return ".json"
    elif content_type == "text/plain":
        return ".txt"
    elif content_type == CONTENT_TYPE_GENERIC:
        return ".dat"
    else:
        candidate = mimetypes.guess_extension(content_type)
        if candidate is not None:
            return candidate
    return ".dat"


def guess_content_type(file_name: str) -> str:
    """
    Guess content type by given file name.

    :param file_name: name of the file
    :return: content type
    """
    return mimetypes.guess_type(file_name)[0] or CONTENT_TYPE_GENERIC


def get_extension(file_name: str) -> str | None:
    """
    Get file extension from file name.

    :param file_name: file name
    :return: file extension with dot (e.g. `.txt`) or None if file doesn't have extension
    """
    parts = file_name.rsplit(".", 1)
    if len(parts) == 2:
        return "." + parts[1]
    return None


def generate_random_file_name(length: int = 40) -> str:
    """
    Generate random string suitable for a file name.

    It doesn't include extension.
    :param length: length of the string to be generated
    :return: random string
    """
    return "".join(random.choice(FILE_NAME_ALPHABET) for _ in range(length))


def ensure_dir(dir_name: os.PathLike | str) -> None:
    """
    Ensure that given directory exists.

    :param dir_name: name of the directory to be ensured
    """
    os.makedirs(dir_name, exist_ok=True)


def calculate_sha1(input_stream: IO | TextIO | IOBase) -> str:
    """
    Calculate SHA1 of the file.

    :param input_stream: input stream to be processed
    :return: checksum in hex string format
    """
    checksum_fn = hashlib.sha1()
    for chunk in iter(lambda: input_stream.read(4096), b""):
        checksum_fn.update(cast(bytes, chunk))
    return checksum_fn.hexdigest()


def calculate_checksum(input_stream: IO | TextIO | IOBase) -> str:
    """
    Calculate checksum of the file.

    :param input_stream: input stream to be processed
    :return: checksum in hex string format
    """
    return calculate_sha1(input_stream)


TIO = TypeVar("TIO", bound=Union[IO, TextIO])
FModeBinary = Literal["rb", "wb", "ab", "rb+", "wb+", "ab+", "xb", "xb+"]
FModeText = Literal["r", "w", "a", "r+", "w+", "a+", "x", "x+"]



#
#  Copyright 2024 by Dmitry Berezovsky, MIT License
#
import abc
from typing import Any


class Abstract:
    """
    Marker class to mark a class as abstract.

    This object makes no sense without the `AbstractMeta` metaclass.
    See the `AbstractMeta` for usage instructions.
    """

    __slots__ = ()


class AbstractMeta(abc.ABCMeta):
    """
    Improved version of the standard ABCMeta metaclass for defining abstract classes.

    This object inherits the standard behavior of the ABCMeta but also allows to define abstract classes
    WITHOUT any ABSTRACT methods. This is particularly useful when you want to override just static fields in the
    descendants.

    Python standard considers class as abstract if it defines at least one abstract method. So if you want just a marker
    interface to be abstract, you have to define a dummy abstract method and then override it in the descendants.

    This meta actually does exactly the same behind the scene - when you mark you class with `Abstract` by adding this
    class into the list of baseclasses it will add a dummy abstract method to the class, if `Abstract` is not in the
    list of direct parents, it will be implemented automatically.

    It is safe to use this metaclass as a drop-in replacement for ABCMeta even if you don't need `Abstract` marker.

    Usage:

    ```python

    class MyAbstractClass(Abstract, metaclass=AbstractMeta):
        pass

    class MyImpl(MyAbstractClass):
        pass

    isabstract(MyAbstractClass)     # Is True
    isabstract(MyImpl)              # Is False
    ```
    """

    def __new__(
        mcls, name: str, bases: tuple[type, ...], namespace: dict[str, Any], /, **kwargs: Any
    ) -> "AbstractMeta":
        """See class description."""
        result = super().__new__(mcls, name, bases, namespace, **kwargs)
        from _abc import _abc_init  # type: ignore

        if Abstract in bases:
            result._abstract_dummy_abstract_method = lambda: None  # type: ignore
            result._abstract_dummy_abstract_method.__isabstractmethod__ = True  # type: ignore
            _abc_init(result)
        elif hasattr(result, "_abstract_dummy_abstract_method"):
            result._abstract_dummy_abstract_method = lambda: None
            _abc_init(result)
        return result



#
#  Copyright 2024 by Dmitry Berezovsky, MIT License
#
import abc
import datetime
import time


class ExecutionTimer(metaclass=abc.ABCMeta):
    """
    Class for measuring execution time of a block of code.

    Example:
        timer = ExecutionTimer()
        try:
            for x in range(5):
                sleep(random())
        finally:
            timer.stop()
        print("Execution finished in " + timer.format_elapsed())
    """

    def __init__(self, start: bool = True) -> None:
        if start:
            self.start()

    @abc.abstractmethod
    def start(self) -> None:
        """Start the timer."""
        pass

    @abc.abstractmethod
    def stop(self) -> None:
        """Stop the timer."""
        pass

    @abc.abstractmethod
    def format_elapsed(self) -> str:
        """Format elapsed time as a string."""
        pass

    @abc.abstractmethod
    def format_elapsed_sec(self) -> str:
        """Format elapsed time in fractional seconds as a string."""
        pass

    @abc.abstractmethod
    def get_eta(self, processed: int, total: int) -> datetime.timedelta | None:
        """Return estimated time of arrival (ETA) for a given number of processed items and total number of items."""
        pass

    @property
    @abc.abstractmethod
    def elapsed(self) -> datetime.timedelta | None:
        """
        Return elapsed time as a timedelta object.

        If timer has not been started returns None.
        :return: timedelta or none
        """
        pass

    @property
    @abc.abstractmethod
    def is_running(self) -> bool:
        """Return True if timer is running, False otherwise."""
        pass

    @property
    @abc.abstractmethod
    def elapsed_sec(self) -> float | None:
        """
        Return elapsed time in fractional seconds.

        If timer has not been started returns None.
        :return: timedelta or none
        """
        pass

    @property
    @abc.abstractmethod
    def is_finished(self) -> bool:
        """Return True if timer has been stopped, False otherwise."""
        pass


class PerfExecutionTimer(ExecutionTimer):
    """
    Implementation of the ExecutionTimer which relies on `time.perf_timer`.

    This method provides accurate and _monotonic_ clock so could be used for very precise short-time measurements.
    """

    def __init__(self, start: bool = True) -> None:
        self.__start_time: float | None = None
        self.__end_time: float | None = None
        super().__init__(start)

    def start(self) -> None:
        """Start the timer."""
        self.__start_time = time.perf_counter()
        self.__end_time = None

    def stop(self) -> None:
        """Stop the timer."""
        self.__end_time = time.perf_counter()

    @property
    def elapsed(self) -> datetime.timedelta | None:
        """
        Return elapsed time as a timedelta object.

        If timer has not been started returns None.
        :return: timedelta or none
        """
        secs = self.elapsed_sec
        if secs is None:
            return None
        return datetime.timedelta(seconds=secs)

    @property
    def is_running(self) -> bool:
        """Return True if timer is running, False otherwise."""
        return self.__start_time is not None and self.__end_time is None

    @property
    def elapsed_sec(self) -> float | None:
        """
        Return elapsed time in fractional seconds.

        If timer has not been started returns None.
        :return: timedelta or none
        """
        if self.is_finished:
            return self.__end_time - self.__start_time  # type: ignore
        if self.is_running:
            return time.perf_counter() - self.__start_time  # type: ignore
        else:
            return None

    @property
    def is_finished(self) -> bool:
        """Return True if timer has been stopped, False otherwise."""
        return self.__start_time is not None and self.__end_time is not None

    def format_elapsed(self) -> str:
        """Format elapsed time as a string."""
        val = self.elapsed
        return str(val) if val is not None else "n/a"

    def format_elapsed_sec(self) -> str:
        """Format elapsed time in fractional seconds as a string."""
        val = self.elapsed_sec
        if val is None:
            return "n/a"
        else:
            return f"{round(val, 2)}s"

    def get_eta(self, processed: int, total: int) -> datetime.timedelta | None:
        """Return estimated time of arrival (ETA) for a given number of processed items and total number of items."""
        assert processed > 0, "Could be used only after some progress is done"
        if self.elapsed_sec is None:
            return None
        return datetime.timedelta(seconds=(total - processed) * self.elapsed_sec / processed)



#
#  Copyright 2024 by Dmitry Berezovsky, MIT License
#
from collections.abc import Collection, Iterable
from typing import Sequence

from .types import T


def filter_none(collection: Sequence[T]) -> list[T]:
    """Return the copy of given list without None elements."""
    return list(filter(None, collection))


def filter_unique(collection: Collection[T]) -> list[T]:
    """Return a copy of passed `collection` without duplicates preserving order."""
    return list(dict.fromkeys(collection))


def ensure_list(list_or_obj: T | Iterable[T]) -> list[T]:
    """Wrap given object into list and returns as result unless given object is not list already."""
    if not isinstance(list_or_obj, (list, tuple)):
        return [list_or_obj]  # type: ignore
    if isinstance(list_or_obj, tuple):
        return list(list_or_obj)
    return list_or_obj



#
#  Copyright 2024 by Dmitry Berezovsky, MIT License
#
import functools
from importlib import import_module
from inspect import isfunction, ismethod
import logging
import threading
from typing import Any, Callable, Sequence, cast, get_type_hints

from django.conf import settings
from django.contrib.sites import management
from django.core.handlers.asgi import ASGIRequest
from django.http import HttpRequest, HttpResponse
from django.urls import URLPattern, URLResolver
from django.views.decorators.csrf import csrf_exempt
from injector import Binder, Injector

from unikit.contrib.django.di.base import BaseDiSupportedApp
from unikit.di import DiModule

logger = logging.getLogger(__name__)


class _DjangoInjectionApp(BaseDiSupportedApp):
    """
    Django DI app enables DI django lifecycle (views, template processors, etc).

    NEVER USE THIS CLASS DIRECTLY. INSTALL application `unikit.contrib.django.di` in INSTALLED_APPS instead.
    """

    ENABLE_VIEW_INJECTION = True
    ENABLE_TEMPLATE_PROCESSOR_INJECTION = True
    ENABLE_COMMAND_INJECTION = True
    default = False

    def __init__(self, app_name: str, app_module: str) -> None:
        super().__init__(app_name, app_module)
        self.django_module = _DjangoDiModule()
        self._di_container.binder.install(self.django_module)


class _DjangoDiModule(DiModule):

    def __init__(self) -> None:
        super().__init__()
        self._local = threading.local()

    def set_request(self, request: HttpRequest) -> None:
        """Set the current request in the thread local storage."""
        if isinstance(request, ASGIRequest):
            if settings.DEBUG:
                logger.warning(
                    "Calling DjangoModule.set_request with a ASGIRequest will lead to "
                    "bad results because the asgi handler does not spawn a thread per "
                    "request. Ignoring call to set_request"
                )
            self._local.request = None
            return
        self._local.request = request

    def get_request(self) -> HttpRequest | None:
        """Return the current request from the thread local storage."""
        try:
            return self._local.request
        except AttributeError:
            return None

    def configure(self, binder: Binder) -> None:
        """Configure the DI module."""
        super().configure(binder)
        from django.conf import LazySettings, Settings, settings

        self.register_singleton(Settings, to=settings)
        self.register_singleton(LazySettings, to=settings)
        binder.bind(HttpRequest, to=lambda: self.get_request())


def patch_command_loader(injector: Injector) -> None:
    """Patches the management command loader to allow injection into management commands."""

    # Original at:
    # https://github.com/django/django/blob/master/django/core/management/__init__.py#L33

    def load_command_class(app_name: str, name: str) -> None:
        module = import_module(f"{app_name}.management.commands.{name}")
        return injector.create_object(cast(Any, module).Command)

    management.load_command_class = load_command_class


def _get_before_request_hook(view_obj: Callable | type, request: HttpRequest) -> HttpResponse | None:
    """
    Return the before request hook if it exists.

    If the view object has attached before request hook it will be invoked before actual view logic.
    Hook may perform any some extra logic and either return None (meaning the normal view logic should be executed) or
    HttpResponse (meaning which means the execution must stop and the generated response must be returned).

    This method extracts the hook, if the hook doesn't exist - returns none, if hook exist launches it and returns hook
    result.

    :param view_obj: view function or class
    :param request: incoming http request
    :return: HtpResponse or none
    """
    hook = getattr(view_obj, "_on_before_request", None)
    if isfunction(hook):
        return hook(request)
    return None


def _get_after_request_hook(view_obj: Callable | type, response: HttpResponse) -> HttpResponse:
    """
    Return after request hook if it exists.

    If the view object has attached after request hook it will be invoked after view logic but before sending the
    http response to the further processing.

    Hook may perform any some extra logic and modify the response. The valid use case is to modify headers.

    This method extracts the hook, if the hook doesn't exist - returns original response,
    if hook exist launches it and returns result of the hook. Hook MUST return the HttpResponse.
    result.

    :param view_obj: view function or class
    :param response: response generated by view or before_request_hook
    :return: HtpResponse or none
    """
    hook = getattr(view_obj, "_on_after_request", None)
    if isfunction(hook):
        return hook(response)
    return response


# Code below is based on open-source project django_injector distributed under BSD-2 License
# and hosted at https://github.com/blubber/django_injector
# Bellow is a copy of original license
#
# BSD 2-Clause License
#
# Copyright (c) 2019, Tiemo
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


def _process_list(lst: Sequence, injector: Injector) -> list:
    return [__wrap_fun(f, injector) for f in lst]


def _process_resolver(resolver: URLResolver, injector: Injector) -> None:
    if resolver.callback:
        resolver.callback = __wrap_fun(resolver.callback, injector)

    for pattern in resolver.url_patterns:
        if isinstance(pattern, URLPattern) and pattern.callback:
            pattern.callback = __wrap_fun(pattern.callback, injector)
        elif isinstance(pattern, URLResolver):
            _process_resolver(pattern, injector)

    if resolver._populated:
        resolver._populate()


def __check_existing_csrf_exempt(fun: Callable, wrapper: Callable) -> Callable:
    if hasattr(fun, "csrf_exempt") and fun.csrf_exempt:
        # Graphene-Django common solution for csrf_exempt is already applied in urls
        wrapper.csrf_exempt = True  # type: ignore
    return wrapper


def __wrap_function(fun: Callable, injector: Injector) -> Callable:
    @functools.wraps(fun)
    def wrapper(*args: Any, **kwargs: Any) -> Any:
        return injector.call_with_injection(callable=fun, args=args, kwargs=kwargs)

    return __check_existing_csrf_exempt(fun, wrapper)


def __instance_method_wrapper(im: Callable) -> Callable:
    @functools.wraps(im)
    def wrapper(*args: Any, **kwargs: Any) -> Any:
        return im(*args, **kwargs)

    return wrapper


def __wrap_class_based_view(fun: Callable, injector: Injector) -> Callable:
    cls = cast(Any, fun).view_class

    try:
        initkwargs = cast(Any, fun).view_initkwargs
    except AttributeError:
        return fun

    # Code copied from Django's django.views.generic.base.View
    # to enable injection into class based view constructors.
    def view(request: Any, *args: Any, **kwargs: Any) -> Any:
        hook_result = _get_before_request_hook(fun, request)
        if hook_result:
            return hook_result

        self = injector.create_object(cls, additional_kwargs=initkwargs)
        injector.call_with_injection(
            callable=self.setup,
            args=(request,) + args,
            kwargs=kwargs,
        )
        if not hasattr(self, "request"):
            raise AttributeError(
                "%s instance has no 'request' attribute. Did you override "
                "setup() and forget to call super()?" % cls.__name__
            )
        return _get_after_request_hook(fun, self.dispatch(request, *args, **kwargs))

    cast(Any, view).view_class = cls
    cast(Any, view).view_initkwargs = initkwargs
    # masking our view for resolver match so we can understand what underlying
    # class-based view is there
    cast(Any, view).__module__ = cls.__module__
    cast(Any, view).__name__ = cls.__name__

    if hasattr(fun, "cls"):
        # Django Rest Framework APIView's as_view returns a callable with a cls attribute
        # instead of view_class. If the original callable has a cls attribute, assume it's
        # DRF. In addition, DRF views are csrf exempt by default, the SessionAuthentication
        # auth backend will selectively apply CSRF protection.
        cast(Any, view).cls = cls
        cast(Any, view).initkwargs = initkwargs
        view = csrf_exempt(view)

    return __check_existing_csrf_exempt(fun, view)


def __wrap_fun(fun: Callable, injector: Injector) -> Callable:
    if ismethod(fun):
        fun = __instance_method_wrapper(fun)

    # This blocks needs to come before the block that checks for __call__
    # to prevent infinite recursion.
    if hasattr(fun, "__bindings__"):
        return __wrap_function(fun, injector)

    if hasattr(fun, "view_class"):
        return __wrap_class_based_view(fun, injector)

    if callable(fun) and not isinstance(fun, type):
        try:
            type_hints = get_type_hints(fun)
        except (AttributeError, TypeError):
            # Some callables are not supported by get_type_hints:
            # https://github.com/alecthomas/flask_injector/blob/master/flask_injector/__init__.py#L75
            wrap_it = False
        else:
            type_hints.pop("return", None)
            wrap_it = type_hints != {}
        if wrap_it:
            return __wrap_function(fun, injector)  # Original: return wrap_fun(inject(fun), injector)

    return fun
